@@ -1,13 +1,13 @@
+from typing_extensions import Final, Protocol
+
 import json
 import logging
 import math
 import time
-from pathlib import Path
-from typing import Callable, Dict, Generic, Iterable, List, Optional, TypeVar
-from typing_extensions import Final, Protocol
-
 import torch
+from pathlib import Path
 from tqdm import tqdm
+from typing import Callable, Dict, Generic, Iterable, List, Optional, TypeVar
 
 from ptgnn.baseneuralmodel.abstractneuralmodel import AbstractNeuralModel
 from ptgnn.baseneuralmodel.modulewithmetrics import ModuleWithMetrics
@@ -51,6 +51,7 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         clip_gradient_norm: Optional[float] = None,
         target_validation_metric: Optional[str] = None,
         target_validation_metric_higher_is_better: bool = False,
+        enable_amp: bool = False,
     ):
         """
         :param model: The Component to be built and trained
@@ -84,6 +85,7 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         self.__train_epoch_end_hooks: List[EndOfEpochHook] = []
         self.__validation_epoch_end_hooks: List[EndOfEpochHook] = []
         self.__clip_gradient_norm = clip_gradient_norm
+        self.__enable_amp = enable_amp
 
         self.__target_metric = target_validation_metric
         if target_validation_metric is not None:
@@ -108,6 +110,13 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
     def neural_module(self, nn: TNeuralModule):
         self.__neural_network = nn
 
+    def load_metadata_and_create_network(
+        self, training_data: Iterable[TRawDatapoint], parallelize: bool, show_progress_bar: bool
+    ) -> None:
+        return self.__load_metadata_and_create_network(
+            training_data, parallelize, show_progress_bar
+        )
+
     def __load_metadata_and_create_network(
         self, training_data: Iterable[TRawDatapoint], parallelize: bool, show_progress_bar: bool
     ) -> None:
@@ -129,10 +138,18 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         self.LOGGER.info(
             "Model metadata loaded. The following model was created:\n %s", self.__neural_network
         )
+
+        for m in self.__metadata_finalized_hooks:
+            m(self.__model)
+
         self.LOGGER.info(
-            "Model Definition:\n %s", json.dumps(dict(self.__model.model_definition), indent=2),
+            "Model Definition:\n %s",
+            json.dumps(dict(self.__model.model_definition), indent=2),
         )
 
+        self.LOGGER.info("Saving model with finalized metadata.")
+        self.__save_checkpoint()
+
     def __save_checkpoint(self) -> None:
         self.__model.save(self.__checkpoint_location, self.neural_module)
 
@@ -165,10 +182,13 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         parallelize,
         scheduler,
         show_progress_bar,
+        shuffle_input: bool = True,
     ):
         sum_epoch_loss, running_avg_loss, num_minibatches, num_samples = 0.0, 0.0, 0, 0
         start_time = time.time()
         self.neural_module.train()
+
+        scaler = torch.cuda.amp.GradScaler(enabled=self.__enable_amp)
         with tqdm(desc="Training", disable=not show_progress_bar, leave=False) as progress_bar:
             for step_idx, (mb_data, raw_samples) in enumerate(
                 self.__model.minibatch_iterator(
@@ -176,25 +196,28 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
                     device=device,
                     max_minibatch_size=self.__minibatch_size,
                     yield_partial_minibatches=False,
-                    shuffle_input=True,
+                    shuffle_input=shuffle_input,
                     parallelize=parallelize,
                 )
             ):
                 optimizer.zero_grad()
-                mb_loss = self.neural_module(**mb_data)
-                mb_loss.backward()
+                with torch.cuda.amp.autocast(enabled=self.__enable_amp):
+                    mb_loss = self.neural_module(**mb_data)
+                    if torch.isnan(mb_loss):
+                        raise Exception("Loss has a NaN value.")
 
-                if self.__clip_gradient_norm is not None:
-                    torch.nn.utils.clip_grad_norm_(
-                        self.neural_module.parameters(recurse=True), self.__clip_gradient_norm
-                    )
+                    scaler.scale(mb_loss).backward()
 
-                if torch.isnan(mb_loss):
-                    raise Exception("Loss has a NaN value.")
+                    if self.__clip_gradient_norm is not None:
+                        scaler.unscale_(optimizer)
+                        torch.nn.utils.clip_grad_norm_(
+                            self.neural_module.parameters(recurse=True), self.__clip_gradient_norm
+                        )
 
-                optimizer.step()
-                if scheduler is not None:
-                    scheduler.step(epoch_idx=epoch, epoch_step=step_idx)
+                    scaler.step(optimizer)
+                    scaler.update()
+                    if scheduler is not None:
+                        scheduler.step(epoch_idx=epoch, epoch_step=step_idx)
 
                 num_minibatches += 1
                 num_samples += len(raw_samples)
@@ -245,7 +268,8 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
                 shuffle_input=False,
                 parallelize=parallelize,
             ):
-                mb_loss = self.neural_module(**mb_data)
+                with torch.cuda.amp.autocast(enabled=self.__enable_amp):
+                    mb_loss = self.neural_module(**mb_data)
                 num_minibatches += 1
                 num_samples += len(raw_samples)
                 sum_epoch_loss += float(mb_loss.cpu())
@@ -295,6 +319,7 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
         exponential_running_average_factor: float = 0.97,
         device=None,
         store_tensorized_data_in_memory: bool = False,
+        shuffle_training_data: bool = True,
     ) -> None:
         """
         The training-validation loop for `AbstractNeuralModel`s.
@@ -312,11 +337,10 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
             displayed in the progress bar.
         :param device: the target PyTorch device for training
         :param store_tensorized_data_in_memory: store all tensorized data in memory instead of computing them on-line.
+        :param shuffle_training_data: shuffle the incoming data from `training_data`.
         """
         if initialize_metadata:
-            self.__load_metadata_and_create_network(training_data, parallelize, show_progress_bar)
-            self.LOGGER.info("Saving model with finalized metadata.")
-            self.__save_checkpoint()
+            self.load_metadata_and_create_network(training_data, parallelize, show_progress_bar)
 
         self.LOGGER.info(
             "Model has %s trainable parameters.",
@@ -372,6 +396,7 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
                 parallelize,
                 scheduler,
                 show_progress_bar,
+                shuffle_training_data,
             )
 
             target_metric, target_metric_improved = self._run_validation(
@@ -400,4 +425,4 @@ class ModelTrainer(Generic[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
                     break
 
         # Restore the best model params that were found.
-        self.__restore_checkpoint()
+        self.__restore_checkpoint(device=device)
